// Copyright(c) 2017-2018, Intel Corporation
//
// Redistribution  and  use  in source  and  binary  forms,  with  or  without
// modification, are permitted provided that the following conditions are met:
//
// * Redistributions of  source code  must retain the  above copyright notice,
//   this list of conditions and the following disclaimer.
// * Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.
// * Neither the name  of Intel Corporation  nor the names of its contributors
//   may be used to  endorse or promote  products derived  from this  software
//   without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,  BUT NOT LIMITED TO,  THE
// IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT  SHALL THE COPYRIGHT OWNER  OR CONTRIBUTORS BE
// LIABLE  FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR
// CONSEQUENTIAL  DAMAGES  (INCLUDING,  BUT  NOT LIMITED  TO,  PROCUREMENT  OF
// SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE,  DATA, OR PROFITS;  OR BUSINESS
// INTERRUPTION)  HOWEVER CAUSED  AND ON ANY THEORY  OF LIABILITY,  WHETHER IN
// CONTRACT,  STRICT LIABILITY,  OR TORT  (INCLUDING NEGLIGENCE  OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
#ifndef _C_TEST_SYSTEM_H
#define _C_TEST_SYSTEM_H

#include <stdio.h>
#include <dirent.h>
#include <sched.h>
#include <glob.h>
#ifdef __cplusplus
extern "C" {
#endif

  typedef int (*filter_func)(const struct dirent *);
  typedef int (*compare_func)(const struct dirent **, const struct dirent **);
  int opae_test_open(const char *path, int flags);
  int opae_test_open_create(const char *path, int flags, mode_t mode);
  ssize_t opae_test_read(int fd, void *buf, size_t count);

  FILE * opae_test_fopen(const char *path, const char *mode);

  FILE * opae_test_popen(const char *cmd, const char *type);
  int opae_test_pclose(FILE *stream);

  int opae_test_close(int fd);
  int opae_test_ioctl(int fd, unsigned long request, va_list argp);

  DIR *opae_test_opendir(const char *name);
  ssize_t opae_test_readlink(const char *path, char *buf, size_t bufsize);
  int opae_test_xstat(int ver, const char *path, struct stat *buf);
  int opae_test_lstat(int ver, const char *path, struct stat *buf);
  int opae_test_scandir(const char *dirp, struct dirent ***namelist, filter_func filter, compare_func cmp);

  int opae_test_sched_setaffinity(pid_t pid, size_t cpusetsize, const cpu_set_t *mask);
  
  int opae_test_glob(const char *pattern, int flags,
                int (*errfunc) (const char *epath, int eerrno),
                glob_t *pglob);
  char *opae_test_realpath(const char *inp, char *dst);

#ifdef __cplusplus
}
#endif

#endif /* !_TEST_SYSTEM_H */
