// Copyright(c) 2018, Intel Corporation
//
// Redistribution  and  use  in source  and  binary  forms,  with  or  without
// modification, are permitted provided that the following conditions are met:
//
// * Redistributions of  source code  must retain the  above copyright notice,
//   this list of conditions and the following disclaimer.
// * Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.
// * Neither the name  of Intel Corporation  nor the names of its contributors
//   may be used to  endorse or promote  products derived  from this  software
//   without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,  BUT NOT LIMITED TO,  THE
// IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT  SHALL THE COPYRIGHT OWNER  OR CONTRIBUTORS BE
// LIABLE  FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR
// CONSEQUENTIAL  DAMAGES  (INCLUDING,  BUT  NOT LIMITED  TO,  PROCUREMENT  OF
// SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE,  DATA, OR PROFITS;  OR BUSINESS
// INTERRUPTION)  HOWEVER CAUSED  AND ON ANY THEORY  OF LIABILITY,  WHETHER IN
// CONTRACT,  STRICT LIABILITY,  OR TORT  (INCLUDING NEGLIGENCE  OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
#include "pyevents.h"
namespace py = pybind11;
using opae::fpga::types::event;

const char *event_doc() {
  return R"opaedoc(
    event object are used to register for OPAE events.
    Currently, the event types are:
      EVENT_ERROR
      EVENT_INTERRUPT
      EVENT_POWER_THERMAL
  )opaedoc";
}

const char *event_doc_register_event() {
  return R"opaedoc(
    Factory method used to create events registered to a particular event type.

    Args:
      handle: A handle to an OPAE resource.
      type(fpga_event_type): The event type to register for. Valid event types are:
                             EVENT_ERROR, EVENT_INTERRUPT, EVENT_POWER_THERMAL
      flags: Event registration flags that control registration behavior.
  )opaedoc";
}

opae::fpga::types::event::ptr_t event_register_event(
    opae::fpga::types::handle::ptr_t hndl, fpga_event_type type, int flags) {
  return event::register_event(hndl, type, flags);
}

const char *event_doc_os_object() {
  return R"opaedoc(
    Get an OS specific object from the event which can be used to subscribe for
    events. On Linux, the object corresponds to a file descriptor that can be
    used with select/poll/epoll calls.
  )opaedoc";
}

int event_os_object(opae::fpga::types::event::ptr_t evnt) {
  return evnt->os_object();
}
